from datetime import datetime
from typing import Union

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import CustomModel, ModelInvocation
from okareo_api_client.models import ScenarioSetResponse, TestRunType
from okareo_api_client.models.datapoint_search import DatapointSearch
from okareo_api_client.models.find_test_data_point_payload import (
    FindTestDataPointPayload,
)
from okareo_api_client.models.full_data_point_item import FullDataPointItem
from okareo_api_client.models.full_data_point_item_metric_value import (
    FullDataPointItemMetricValue,
)
from okareo_api_client.models.scenario_data_poin_response import (
    ScenarioDataPoinResponse,
)
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.test_data_point_item import TestDataPointItem
from okareo_api_client.models.test_data_point_item_metric_value import (
    TestDataPointItemMetricValue,
)

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")
rnd_str = random_string(5)
unique_key = f"{rnd_str} {today_with_time}"
create_scenario_name = f"ci_json_test_get_data_points {unique_key}"


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


JSON_SCENARIO = [
    {
        "input": {"animal": "fish", "color": "red"},
        "result": "red",
    },
    {
        "input": {"animal": "dog", "color": "blue"},
        "result": "blue",
    },
    {
        "input": {"animal": "cat", "color": "green"},
        "result": "green",
    },
]  # type: ignore
SCENARIO_INPUTS = [scenario["input"]["animal"] for scenario in JSON_SCENARIO]  # type: ignore
SCENARIO_RESULTS = [scenario["result"] for scenario in JSON_SCENARIO]  # type: ignore

JSON_SEED = Okareo.seed_data_from_list(JSON_SCENARIO)  # type: ignore


@pytest.fixture(scope="module")
def create_scenario_set(okareo_client: Okareo) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=create_scenario_name,
        seed_data=JSON_SEED,
    )
    scenario = okareo_client.create_scenario_set(scenario_set_create)
    return scenario


def test_get_data_points(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> None:
    test_run_name = f"ci_test_get_data_points {unique_key}"

    class ClassificationModel(CustomModel):
        def invoke(self, input_value: Union[dict, list, str]) -> ModelInvocation:
            assert isinstance(input_value, dict)

            return ModelInvocation(
                model_prediction=input_value["color"], model_output_metadata=input_value
            )

    model_under_test = okareo_client.register_model(
        name=test_run_name,
        model=ClassificationModel(name=test_run_name),
    )

    test_run = model_under_test.run_test(
        scenario=create_scenario_set,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics=True,
    )

    assert test_run.id
    assert test_run.model_metrics
    metrics_dict = test_run.model_metrics.to_dict()
    assert metrics_dict["weighted_average"] is not None
    assert metrics_dict["weighted_average"]["f1"] == 1

    dp = okareo_client.find_datapoints(DatapointSearch(test_run_id=test_run.id))
    assert isinstance(dp, list)
    assert len(dp) == 3
    for d in dp:
        assert isinstance(d.input_, dict)
        assert d.input_["animal"] in SCENARIO_INPUTS
        assert d.input_["color"] in SCENARIO_RESULTS
        assert d.result in SCENARIO_RESULTS
    assert isinstance(create_scenario_set.scenario_id, str)

    sdp = okareo_client.get_scenario_data_points(
        scenario_id=create_scenario_set.scenario_id
    )
    assert isinstance(sdp, list)
    assert len(sdp) == 3
    for sd in sdp:
        assert isinstance(sd, ScenarioDataPoinResponse)
        assert isinstance(sd.input_, dict)
        assert sd.input_["animal"] in SCENARIO_INPUTS
        assert sd.input_["color"] in SCENARIO_RESULTS
        assert sd.result in SCENARIO_RESULTS

    scenario_ids = [s.id for s in sdp]

    tdp = okareo_client.find_test_data_points(
        FindTestDataPointPayload(test_run_id=test_run.id)
    )
    assert isinstance(tdp, list)
    assert len(dp) == 3
    for td in tdp:
        assert isinstance(td, TestDataPointItem) or isinstance(td, FullDataPointItem)
        assert td.scenario_data_point_id in scenario_ids
        assert td.test_run_id == test_run.id
        assert td.metric_type == "MULTI_CLASS_CLASSIFICATION"
        assert isinstance(td.metric_value, TestDataPointItemMetricValue) or isinstance(
            td.metric_value, FullDataPointItemMetricValue
        )
        assert (
            td.metric_value.additional_properties["expected"]
            == td.metric_value.additional_properties["actual"]
        )



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.check_create_update_schema_check_config import CheckCreateUpdateSchemaCheckConfig


T = TypeVar("T", bound="CheckCreateUpdateSchema")


@_attrs_define
class CheckCreateUpdateSchema:
    """
    Attributes:
        name (Union[Unset, str]):
        description (Union[Unset, str]):
        check_config (Union[Unset, CheckCreateUpdateSchemaCheckConfig]):
        project_id (Union[Unset, str]): ID of the project
    """

    name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    check_config: Union[Unset, "CheckCreateUpdateSchemaCheckConfig"] = UNSET
    project_id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        description = self.description
        check_config: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.check_config, Unset):
            check_config = self.check_config.to_dict()

        project_id = self.project_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description
        if check_config is not UNSET:
            field_dict["check_config"] = check_config
        if project_id is not UNSET:
            field_dict["project_id"] = project_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.check_create_update_schema_check_config import CheckCreateUpdateSchemaCheckConfig

        d = src_dict.copy()
        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        _check_config = d.pop("check_config", UNSET)
        check_config: Union[Unset, CheckCreateUpdateSchemaCheckConfig]
        if isinstance(_check_config, Unset):
            check_config = UNSET
        else:
            check_config = CheckCreateUpdateSchemaCheckConfig.from_dict(_check_config)

        project_id = d.pop("project_id", UNSET)

        check_create_update_schema = cls(
            name=name,
            description=description,
            check_config=check_config,
            project_id=project_id,
        )

        check_create_update_schema.additional_properties = d
        return check_create_update_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.error_response import ErrorResponse
from ...models.get_filter_group_notifications_v0_filter_group_notifications_get_response_200_item import (
    GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item,
)
from ...types import Response


def _get_kwargs(
    *,
    api_key: str,
) -> Dict[str, Any]:
    headers = {}
    headers["api-key"] = api_key

    return {
        "method": "get",
        "url": "/v0/filter_group_notifications",
        "headers": headers,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item.from_dict(
                response_200_item_data
            )

            response_200.append(response_200_item)

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ErrorResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    api_key: str,
) -> Response[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    """Get Filter Group Notifications

     Get all notification settings for filter groups in the current project.

    This endpoint retrieves information about all filter groups with notification
    configurations, including their enabled status and channel types (email/slack).

    Args:
        request: The HTTP request

    Returns:
        list: A list of dictionaries containing notification settings for each filter group

    Args:
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, List['GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item']]]
    """

    kwargs = _get_kwargs(
        api_key=api_key,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    api_key: str,
) -> Optional[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    """Get Filter Group Notifications

     Get all notification settings for filter groups in the current project.

    This endpoint retrieves information about all filter groups with notification
    configurations, including their enabled status and channel types (email/slack).

    Args:
        request: The HTTP request

    Returns:
        list: A list of dictionaries containing notification settings for each filter group

    Args:
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, List['GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item']]
    """

    return sync_detailed(
        client=client,
        api_key=api_key,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    api_key: str,
) -> Response[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    """Get Filter Group Notifications

     Get all notification settings for filter groups in the current project.

    This endpoint retrieves information about all filter groups with notification
    configurations, including their enabled status and channel types (email/slack).

    Args:
        request: The HTTP request

    Returns:
        list: A list of dictionaries containing notification settings for each filter group

    Args:
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, List['GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item']]]
    """

    kwargs = _get_kwargs(
        api_key=api_key,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    api_key: str,
) -> Optional[Union[ErrorResponse, List["GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item"]]]:
    """Get Filter Group Notifications

     Get all notification settings for filter groups in the current project.

    This endpoint retrieves information about all filter groups with notification
    configurations, including their enabled status and channel types (email/slack).

    Args:
        request: The HTTP request

    Returns:
        list: A list of dictionaries containing notification settings for each filter group

    Args:
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, List['GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item']]
    """

    return (
        await asyncio_detailed(
            client=client,
            api_key=api_key,
        )
    ).parsed



from datetime import datetime
from typing import Union

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import CustomModel, ModelInvocation
from okareo_api_client.models import ScenarioSetResponse, ScenarioType, TestRunType
from okareo_api_client.models.generation_tone import GenerationTone
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.scenario_set_generate import ScenarioSetGenerate

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")
rnd_str = random_string(5)
unique_key = f"{rnd_str} {today_with_time}"
create_scenario_name = f"ci_scenario_template_create {unique_key}"
generate_scenario_name = f"ci_scenario_template_generate {unique_key}"


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


JSON_SCENARIO = [
    {
        "input": {
            "document": "WebBizz is dedicated to providing our customers with a seamless online shopping experience. Our platform is designed with user-friendly interfaces to help you browse and select the best products suitable for your needs. We offer a wide range of products from top brands and new entrants, ensuring diversity and quality in our offerings. Our 24/7 customer support is ready to assist you with any queries, from product details, shipping timelines, to payment methods. We also have a dedicated FAQ section addressing common concerns. Always ensure you are logged in to enjoy personalized product recommendations and faster checkout processes.",
            "metadata": {"version": "1", "last update": "2024-05-16T14:15:30Z"},
        },
        "result": "75eaa363-dfcc-499f-b2af-1407b43cb133",
    },
    {
        "input": {
            "document": "Safety and security of your data is our top priority at WebBizz. Our platform employs state-of-the-art encryption methods ensuring your personal and financial information remains confidential. Our two-factor authentication at checkout provides an added layer of security. We understand the importance of timely deliveries, hence we've partnered with reliable logistics partners ensuring your products reach you in pristine condition. In case of any delays or issues, our tracking tool can provide real-time updates on your product's location. We believe in transparency and guarantee no hidden fees or charges during your purchase journey.",
            "metadata": {"version": "2", "last update": "2024-05-16T14:15:30Z"},
        },
        "result": "ac0d464c-f673-44b8-8195-60c965e47525",
    },
    {
        "input": {
            "document": "At WebBizz, we value our customer's feedback and are always striving to improve. Our product review section allows customers to share their experiences and helps others make informed decisions. If unsatisfied with a purchase, our easy return and refund policy ensures hassle-free returns within 30 days of purchase. We also offer a loyalty program, WebBizz Rewards, where customers can earn points with each purchase and avail exclusive discounts. For any further assistance, our 'Live Chat' feature connects you instantly with a customer support representative. We thank you for choosing WebBizz and look forward to serving you again!",
            "metadata": {"version": "3", "last update": "2024-05-16T14:15:30Z"},
        },
        "result": "35a4fd5b-453e-4ca6-9536-f20db7303344",
    },
]

JSON_SEED = Okareo.seed_data_from_list(JSON_SCENARIO)  # type: ignore


@pytest.fixture(scope="module")
def create_scenario_set(okareo_client: Okareo) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=create_scenario_name,
        seed_data=JSON_SEED,
    )
    scenario = okareo_client.create_scenario_set(scenario_set_create)
    return scenario


@pytest.fixture(scope="module")
def generate_scenarios(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    questions: ScenarioSetResponse = okareo_client.generate_scenario_set(
        ScenarioSetGenerate(
            name=generate_scenario_name,
            source_scenario_id=create_scenario_set.scenario_id,
            number_examples=2,
            generation_type=ScenarioType.TEXT_REVERSE_QUESTION,
            generation_tone=GenerationTone.INFORMAL,
            pre_template="""{input.document}""",
            post_template="""{"question": "{generation.input}", "metadata": {input.metadata}}""",
        )
    )

    return questions


def test_generate_scenarios(
    okareo_client: Okareo, generate_scenarios: ScenarioSetResponse
) -> None:
    assert generate_scenarios is not None
    assert generate_scenarios.name == generate_scenario_name
    assert generate_scenarios.scenario_id
    assert generate_scenarios.project_id
    assert generate_scenarios.time_created

    scenario_dps = okareo_client.get_scenario_data_points(
        generate_scenarios.scenario_id
    )
    assert scenario_dps
    assert len(scenario_dps) == 6  # 3 documents * 2 questions each
    for dp in scenario_dps:
        input_ = dp.to_dict()["input"]
        result = dp.to_dict()["result"]

        assert input_
        assert result
        assert input_["question"]
        assert input_["metadata"]

        result_to_version = {
            entry["result"]: entry["input"]["metadata"]["version"]  # type: ignore
            for entry in JSON_SCENARIO
        }

        assert input_["metadata"]["version"] == result_to_version[result[0]]


def test_custom_retrieval(
    okareo_client: Okareo, generate_scenarios: ScenarioSetResponse
) -> None:
    test_run_name = f"ci_scenario_template_custom {unique_key}"

    class RetrievalModel(CustomModel):
        def invoke(self, input_value: Union[dict, list, str]) -> ModelInvocation:
            assert isinstance(input_value, dict)
            assert input_value["question"]
            assert input_value["metadata"]

            return ModelInvocation(
                model_prediction=[
                    {
                        "id": "red",
                        "score": 1,
                        "label": "red",
                        "metadata": input_value["metadata"],
                    }
                ],
                model_input=input_value["question"],
                model_output_metadata=input_value["metadata"],
            )

    model_under_test = okareo_client.register_model(
        name=test_run_name,
        model=RetrievalModel(name=test_run_name),
    )

    retrieval_test_run = model_under_test.run_test(
        scenario=generate_scenarios,
        name=test_run_name,
        test_run_type=TestRunType.INFORMATION_RETRIEVAL,
        calculate_metrics=True,
    )

    assert retrieval_test_run.id
    assert retrieval_test_run.model_metrics
    metrics_dict = retrieval_test_run.model_metrics.to_dict()
    assert metrics_dict["Accuracy@k"]



from okareo.checks import CodeBasedCheck


class Check(CodeBasedCheck):
    @staticmethod
    def evaluate(model_output: str, scenario_input: str, scenario_result: str, metadata: dict, model_input: str) -> bool:  # fmt: skip
        # Check if metadata is empty
        if scenario_input != "dummyinput":
            raise ValueError("Check errored out.")
        return True



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.datapoint_summary_item_user_metadata_type_0 import DatapointSummaryItemUserMetadataType0
    from ..models.feedback_range_summary import FeedbackRangeSummary


T = TypeVar("T", bound="DatapointSummaryItem")


@_attrs_define
class DatapointSummaryItem:
    """
    Attributes:
        group (str): The group name or datetime for the data summary.
        datapoints (int): The total count of datapoints for the group.
        issues (int): The total count of issues for the group.
        errors (int): The total count of errors for the group.
        feedback_ranges (Union[Unset, List['FeedbackRangeSummary']]): List of feedback range summaries for the date.
        avg_latency (Union[Unset, float]): The average latency for the group.
        sum_cost (Union[Unset, float]): The total cost for the group.
        user_metadata (Union['DatapointSummaryItemUserMetadataType0', Any, Unset]): Number of distinct values found for
            each key in user metadata for the group.
    """

    group: str
    datapoints: int
    issues: int
    errors: int
    feedback_ranges: Union[Unset, List["FeedbackRangeSummary"]] = UNSET
    avg_latency: Union[Unset, float] = UNSET
    sum_cost: Union[Unset, float] = UNSET
    user_metadata: Union["DatapointSummaryItemUserMetadataType0", Any, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.datapoint_summary_item_user_metadata_type_0 import DatapointSummaryItemUserMetadataType0

        group = self.group
        datapoints = self.datapoints
        issues = self.issues
        errors = self.errors
        feedback_ranges: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.feedback_ranges, Unset):
            feedback_ranges = []
            for feedback_ranges_item_data in self.feedback_ranges:
                feedback_ranges_item = feedback_ranges_item_data.to_dict()

                feedback_ranges.append(feedback_ranges_item)

        avg_latency = self.avg_latency
        sum_cost = self.sum_cost
        user_metadata: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.user_metadata, Unset):
            user_metadata = UNSET

        elif isinstance(self.user_metadata, DatapointSummaryItemUserMetadataType0):
            user_metadata = UNSET
            if not isinstance(self.user_metadata, Unset):
                user_metadata = self.user_metadata.to_dict()

        else:
            user_metadata = self.user_metadata

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "group": group,
                "datapoints": datapoints,
                "issues": issues,
                "errors": errors,
            }
        )
        if feedback_ranges is not UNSET:
            field_dict["feedback_ranges"] = feedback_ranges
        if avg_latency is not UNSET:
            field_dict["avg_latency"] = avg_latency
        if sum_cost is not UNSET:
            field_dict["sum_cost"] = sum_cost
        if user_metadata is not UNSET:
            field_dict["user_metadata"] = user_metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.datapoint_summary_item_user_metadata_type_0 import DatapointSummaryItemUserMetadataType0
        from ..models.feedback_range_summary import FeedbackRangeSummary

        d = src_dict.copy()
        group = d.pop("group")

        datapoints = d.pop("datapoints")

        issues = d.pop("issues")

        errors = d.pop("errors")

        feedback_ranges = []
        _feedback_ranges = d.pop("feedback_ranges", UNSET)
        for feedback_ranges_item_data in _feedback_ranges or []:
            feedback_ranges_item = FeedbackRangeSummary.from_dict(feedback_ranges_item_data)

            feedback_ranges.append(feedback_ranges_item)

        avg_latency = d.pop("avg_latency", UNSET)

        sum_cost = d.pop("sum_cost", UNSET)

        def _parse_user_metadata(data: object) -> Union["DatapointSummaryItemUserMetadataType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _user_metadata_type_0 = data
                user_metadata_type_0: Union[Unset, DatapointSummaryItemUserMetadataType0]
                if isinstance(_user_metadata_type_0, Unset):
                    user_metadata_type_0 = UNSET
                else:
                    user_metadata_type_0 = DatapointSummaryItemUserMetadataType0.from_dict(_user_metadata_type_0)

                return user_metadata_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointSummaryItemUserMetadataType0", Any, Unset], data)

        user_metadata = _parse_user_metadata(d.pop("user_metadata", UNSET))

        datapoint_summary_item = cls(
            group=group,
            datapoints=datapoints,
            issues=issues,
            errors=errors,
            feedback_ranges=feedback_ranges,
            avg_latency=avg_latency,
            sum_cost=sum_cost,
            user_metadata=user_metadata,
        )

        datapoint_summary_item.additional_properties = d
        return datapoint_summary_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.model_under_test_schema_models import ModelUnderTestSchemaModels


T = TypeVar("T", bound="ModelUnderTestSchema")


@_attrs_define
class ModelUnderTestSchema:
    """
    Attributes:
        name (Union[Unset, str]): Name of the model
        models (Union[Unset, ModelUnderTestSchemaModels]): Models to be added for testing
        sensitive_fields (Union[Unset, List[str]]): List of sensitive fields in the model to redact. Should include
            nested fields, e.g. ['headers.authorization', 'body.password']
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter models in the Okareo app
        project_id (Union[Unset, str]): ID of the project
        update (Union[Unset, bool]): If set to true, the model will be updated instead of returning the existing model
    """

    name: Union[Unset, str] = UNSET
    models: Union[Unset, "ModelUnderTestSchemaModels"] = UNSET
    sensitive_fields: Union[Unset, List[str]] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    project_id: Union[Unset, str] = UNSET
    update: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        models: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.models, Unset):
            models = self.models.to_dict()

        sensitive_fields: Union[Unset, List[str]] = UNSET
        if not isinstance(self.sensitive_fields, Unset):
            sensitive_fields = self.sensitive_fields

        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        project_id = self.project_id
        update = self.update

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if name is not UNSET:
            field_dict["name"] = name
        if models is not UNSET:
            field_dict["models"] = models
        if sensitive_fields is not UNSET:
            field_dict["sensitive_fields"] = sensitive_fields
        if tags is not UNSET:
            field_dict["tags"] = tags
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if update is not UNSET:
            field_dict["update"] = update

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.model_under_test_schema_models import ModelUnderTestSchemaModels

        d = src_dict.copy()
        name = d.pop("name", UNSET)

        _models = d.pop("models", UNSET)
        models: Union[Unset, ModelUnderTestSchemaModels]
        if isinstance(_models, Unset):
            models = UNSET
        else:
            models = ModelUnderTestSchemaModels.from_dict(_models)

        sensitive_fields = cast(List[str], d.pop("sensitive_fields", UNSET))

        tags = cast(List[str], d.pop("tags", UNSET))

        project_id = d.pop("project_id", UNSET)

        update = d.pop("update", UNSET)

        model_under_test_schema = cls(
            name=name,
            models=models,
            sensitive_fields=sensitive_fields,
            tags=tags,
            project_id=project_id,
            update=update,
        )

        model_under_test_schema.additional_properties = d
        return model_under_test_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.full_data_point_item_metric_value import FullDataPointItemMetricValue
    from ..models.full_data_point_item_model_metadata_type_0 import FullDataPointItemModelMetadataType0
    from ..models.full_data_point_item_scenario_input_type_0 import FullDataPointItemScenarioInputType0
    from ..models.full_data_point_item_scenario_result_type_0 import FullDataPointItemScenarioResultType0


T = TypeVar("T", bound="FullDataPointItem")


@_attrs_define
class FullDataPointItem:
    """
    Attributes:
        id (str):
        test_run_id (str):
        metric_type (str):
        metric_value (FullDataPointItemMetricValue):
        tags (Union[Unset, List[str]]):
        scenario_data_point_id (Union[Unset, str]):
        model_input (Union[Unset, Any]):
        model_result (Union[Unset, Any]):
        scenario_input (Union['FullDataPointItemScenarioInputType0', List[Any], Unset, str]):
        scenario_result (Union['FullDataPointItemScenarioResultType0', List[Any], Unset, str]):
        model_metadata (Union['FullDataPointItemModelMetadataType0', Any, Unset]):
        time_created (Union[Unset, str]):
        checks (Union[Unset, Any]):
        end_time (Union[Unset, Any]):
        driver_prompt (Union[Unset, str]):
    """

    id: str
    test_run_id: str
    metric_type: str
    metric_value: "FullDataPointItemMetricValue"
    tags: Union[Unset, List[str]] = UNSET
    scenario_data_point_id: Union[Unset, str] = UNSET
    model_input: Union[Unset, Any] = UNSET
    model_result: Union[Unset, Any] = UNSET
    scenario_input: Union["FullDataPointItemScenarioInputType0", List[Any], Unset, str] = UNSET
    scenario_result: Union["FullDataPointItemScenarioResultType0", List[Any], Unset, str] = UNSET
    model_metadata: Union["FullDataPointItemModelMetadataType0", Any, Unset] = UNSET
    time_created: Union[Unset, str] = UNSET
    checks: Union[Unset, Any] = UNSET
    end_time: Union[Unset, Any] = UNSET
    driver_prompt: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.full_data_point_item_model_metadata_type_0 import FullDataPointItemModelMetadataType0
        from ..models.full_data_point_item_scenario_input_type_0 import FullDataPointItemScenarioInputType0
        from ..models.full_data_point_item_scenario_result_type_0 import FullDataPointItemScenarioResultType0

        id = self.id
        test_run_id = self.test_run_id
        metric_type = self.metric_type
        metric_value = self.metric_value.to_dict()

        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        scenario_data_point_id = self.scenario_data_point_id
        model_input = self.model_input
        model_result = self.model_result
        scenario_input: Union[Dict[str, Any], List[Any], Unset, str]
        if isinstance(self.scenario_input, Unset):
            scenario_input = UNSET

        elif isinstance(self.scenario_input, FullDataPointItemScenarioInputType0):
            scenario_input = UNSET
            if not isinstance(self.scenario_input, Unset):
                scenario_input = self.scenario_input.to_dict()

        elif isinstance(self.scenario_input, list):
            scenario_input = UNSET
            if not isinstance(self.scenario_input, Unset):
                scenario_input = self.scenario_input

        else:
            scenario_input = self.scenario_input

        scenario_result: Union[Dict[str, Any], List[Any], Unset, str]
        if isinstance(self.scenario_result, Unset):
            scenario_result = UNSET

        elif isinstance(self.scenario_result, FullDataPointItemScenarioResultType0):
            scenario_result = UNSET
            if not isinstance(self.scenario_result, Unset):
                scenario_result = self.scenario_result.to_dict()

        elif isinstance(self.scenario_result, list):
            scenario_result = UNSET
            if not isinstance(self.scenario_result, Unset):
                scenario_result = self.scenario_result

        else:
            scenario_result = self.scenario_result

        model_metadata: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.model_metadata, Unset):
            model_metadata = UNSET

        elif isinstance(self.model_metadata, FullDataPointItemModelMetadataType0):
            model_metadata = UNSET
            if not isinstance(self.model_metadata, Unset):
                model_metadata = self.model_metadata.to_dict()

        else:
            model_metadata = self.model_metadata

        time_created = self.time_created
        checks = self.checks
        end_time = self.end_time
        driver_prompt = self.driver_prompt

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "test_run_id": test_run_id,
                "metric_type": metric_type,
                "metric_value": metric_value,
            }
        )
        if tags is not UNSET:
            field_dict["tags"] = tags
        if scenario_data_point_id is not UNSET:
            field_dict["scenario_data_point_id"] = scenario_data_point_id
        if model_input is not UNSET:
            field_dict["model_input"] = model_input
        if model_result is not UNSET:
            field_dict["model_result"] = model_result
        if scenario_input is not UNSET:
            field_dict["scenario_input"] = scenario_input
        if scenario_result is not UNSET:
            field_dict["scenario_result"] = scenario_result
        if model_metadata is not UNSET:
            field_dict["model_metadata"] = model_metadata
        if time_created is not UNSET:
            field_dict["time_created"] = time_created
        if checks is not UNSET:
            field_dict["checks"] = checks
        if end_time is not UNSET:
            field_dict["end_time"] = end_time
        if driver_prompt is not UNSET:
            field_dict["driver_prompt"] = driver_prompt

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.full_data_point_item_metric_value import FullDataPointItemMetricValue
        from ..models.full_data_point_item_model_metadata_type_0 import FullDataPointItemModelMetadataType0
        from ..models.full_data_point_item_scenario_input_type_0 import FullDataPointItemScenarioInputType0
        from ..models.full_data_point_item_scenario_result_type_0 import FullDataPointItemScenarioResultType0

        d = src_dict.copy()
        id = d.pop("id")

        test_run_id = d.pop("test_run_id")

        metric_type = d.pop("metric_type")

        metric_value = FullDataPointItemMetricValue.from_dict(d.pop("metric_value"))

        tags = cast(List[str], d.pop("tags", UNSET))

        scenario_data_point_id = d.pop("scenario_data_point_id", UNSET)

        model_input = d.pop("model_input", UNSET)

        model_result = d.pop("model_result", UNSET)

        def _parse_scenario_input(data: object) -> Union["FullDataPointItemScenarioInputType0", List[Any], Unset, str]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _scenario_input_type_0 = data
                scenario_input_type_0: Union[Unset, FullDataPointItemScenarioInputType0]
                if isinstance(_scenario_input_type_0, Unset):
                    scenario_input_type_0 = UNSET
                else:
                    scenario_input_type_0 = FullDataPointItemScenarioInputType0.from_dict(_scenario_input_type_0)

                return scenario_input_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                scenario_input_type_1 = cast(List[Any], data)

                return scenario_input_type_1
            except:  # noqa: E722
                pass
            return cast(Union["FullDataPointItemScenarioInputType0", List[Any], Unset, str], data)

        scenario_input = _parse_scenario_input(d.pop("scenario_input", UNSET))

        def _parse_scenario_result(
            data: object,
        ) -> Union["FullDataPointItemScenarioResultType0", List[Any], Unset, str]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _scenario_result_type_0 = data
                scenario_result_type_0: Union[Unset, FullDataPointItemScenarioResultType0]
                if isinstance(_scenario_result_type_0, Unset):
                    scenario_result_type_0 = UNSET
                else:
                    scenario_result_type_0 = FullDataPointItemScenarioResultType0.from_dict(_scenario_result_type_0)

                return scenario_result_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                scenario_result_type_1 = cast(List[Any], data)

                return scenario_result_type_1
            except:  # noqa: E722
                pass
            return cast(Union["FullDataPointItemScenarioResultType0", List[Any], Unset, str], data)

        scenario_result = _parse_scenario_result(d.pop("scenario_result", UNSET))

        def _parse_model_metadata(data: object) -> Union["FullDataPointItemModelMetadataType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _model_metadata_type_0 = data
                model_metadata_type_0: Union[Unset, FullDataPointItemModelMetadataType0]
                if isinstance(_model_metadata_type_0, Unset):
                    model_metadata_type_0 = UNSET
                else:
                    model_metadata_type_0 = FullDataPointItemModelMetadataType0.from_dict(_model_metadata_type_0)

                return model_metadata_type_0
            except:  # noqa: E722
                pass
            return cast(Union["FullDataPointItemModelMetadataType0", Any, Unset], data)

        model_metadata = _parse_model_metadata(d.pop("model_metadata", UNSET))

        time_created = d.pop("time_created", UNSET)

        checks = d.pop("checks", UNSET)

        end_time = d.pop("end_time", UNSET)

        driver_prompt = d.pop("driver_prompt", UNSET)

        full_data_point_item = cls(
            id=id,
            test_run_id=test_run_id,
            metric_type=metric_type,
            metric_value=metric_value,
            tags=tags,
            scenario_data_point_id=scenario_data_point_id,
            model_input=model_input,
            model_result=model_result,
            scenario_input=scenario_input,
            scenario_result=scenario_result,
            model_metadata=model_metadata,
            time_created=time_created,
            checks=checks,
            end_time=end_time,
            driver_prompt=driver_prompt,
        )

        full_data_point_item.additional_properties = d
        return full_data_point_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os

DEFAULT_BASE_URL = "https://api.okareo.com"
BASE_URL = os.environ.get("OKAREO_BASE_URL") or os.environ.get(
    "BASE_URL", DEFAULT_BASE_URL
)

DEFAULT_HTTPX_TIME_OUT = 30
# being generous to support the longer generations
HTTPX_TIME_OUT = float(os.environ.get("HTTPX_TIME_OUT", DEFAULT_HTTPX_TIME_OUT))


class NotJSONError(Exception):
    def __init__(self) -> None:
        super().__init__("Expected JSON response, received non-JSON.")



import os

import pytest
from okareo_tests.common import API_KEY, random_string
from okareo_tests.utils import assert_metrics

from okareo import Okareo
from okareo.model_under_test import GenerationModel
from okareo_api_client.models import (
    ScenarioSetCreate,
    ScenarioSetResponse,
    SeedData,
    TestRunType,
)


@pytest.fixture(scope="module")
def rnd() -> str:
    return random_string(5)


@pytest.fixture(scope="module")
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


@pytest.fixture(scope="module")
def article_scenario_set(rnd: str, okareo: Okareo) -> ScenarioSetResponse:
    seed_data = [
        SeedData(
            input_="At WebBizz, we value our customer's feedback and are always striving to improve."
            + "Our product review section allows customers to share their experiences."
            + "This helps others make informed decisions when shopping on our platform.",
            result="35a4fd5b-453e-4ca6-9536-f20db7303344",
        )
    ]
    scenario_set_create = ScenarioSetCreate(
        name=f"ci-test-generate-scenario-set-{rnd}",
        seed_data=seed_data,
    )
    articles: ScenarioSetResponse = okareo.create_scenario_set(scenario_set_create)

    return articles


def run_generation_model_test(
    rnd: str,
    okareo: Okareo,
    article_scenario_set: ScenarioSetResponse,
    model_id: str,
    api_key: str,
    system_prompt: str,
    user_prompt: str,
) -> None:
    mut = okareo.register_model(
        name=f"ci-generation-model-{model_id}-test-{rnd}",
        model=GenerationModel(
            model_id=model_id,
            temperature=0,
            system_prompt_template=system_prompt,
            user_prompt_template=user_prompt,
        ),
    )

    run_resp = mut.run_test(
        name=f"ci-generation-model-{model_id}-run-{rnd}",
        scenario=article_scenario_set,
        api_key=api_key,
        test_run_type=TestRunType.NL_GENERATION,
        checks=["latency"],
    )

    assert run_resp.name == f"ci-generation-model-{model_id}-run-{rnd}"
    assert_metrics(run_resp, num_rows=1, custom_dimensions=["latency"])


def test_gpt35(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    run_generation_model_test(
        rnd,
        okareo,
        article_scenario_set,
        model_id="gpt-4.1-mini",
        api_key=os.environ["OPENAI_API_KEY"],
        system_prompt="You are an AI assistant with a flair for dramatic storytelling and exaggeration.",
        user_prompt="Transform this mundane topic into a very brief dramatic story of heroic proportions: {scenario_input}",
    )


def test_claude3(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    run_generation_model_test(
        rnd,
        okareo,
        article_scenario_set,
        model_id="claude-3-haiku-20240307",
        api_key=os.environ["ANTHROPIC_API_KEY"],
        system_prompt="You are an AI assistant with a flair for dramatic storytelling and exaggeration.",
        user_prompt="Transform this mundane topic into a very brief dramatic story of heroic proportions: {scenario_input}",
    )


def test_cohere(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    run_generation_model_test(
        rnd,
        okareo,
        article_scenario_set,
        model_id="command-r",
        api_key=os.environ["COHERE_API_KEY"],
        system_prompt="You are an AI assistant with a flair for dramatic storytelling and exaggeration.",
        user_prompt="Transform this mundane topic into a very brief dramatic story of heroic proportions: {scenario_input}",
    )



from pytest_httpx import HTTPXMock

from okareo import Okareo


def test_register_model(httpx_mock: HTTPXMock) -> None:
    fixture = {
        "id": "1",
        "project_id": "1",
        "name": "NotebookModel",
        "tags": ["ci-testing"],
        "time_created": "foo",
    }
    httpx_mock.add_response(
        json=[
            {
                "id": "0156f5d7-4ac4-4568-9d44-24750aa08d1a",
                "name": "Global",
                "onboarding_status": "onboarding_status",
                "tags": [],
                "additional_properties": {},
            }
        ],
        status_code=201,
    )
    httpx_mock.add_response(status_code=201, json=fixture)
    okareo = Okareo("api-key", "http://mocked.com")
    mut = okareo.register_model(name="NotebookModel", tags=["ci-testing"])
    assert mut



import pytest
from okareo_tests.checks.sample_check import Check
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo_api_client.models import EvaluatorSpecRequest


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


def test_get_all_checks(okareo_client: Okareo) -> None:
    checks = okareo_client.get_all_checks()
    # iterate through all the checks and ensure they have required fields
    for check in checks:
        assert check.id
        assert check.name
        assert type(check.description) is str
        assert type(check.output_data_type) is str
        assert check.time_created
        assert isinstance(check.is_predefined, bool)
        # concurrency issues with this test when ran in parallel


def test_generate_and_create_check(okareo_client: Okareo) -> None:
    generate_request = EvaluatorSpecRequest(
        description="""
        Return True if the model_output is at least 20 characters long, otherwise return False.""",
        requires_scenario_input=False,
        requires_scenario_result=False,
        output_data_type="bool",
    )
    check = okareo_client.generate_check(generate_request)
    assert check.generated_code
    uploaded_check = okareo_client.create_or_update_check(
        name=f"test_create_check {random_string(5)}",
        description="Test check",
        check=Check(),
    )
    assert uploaded_check.id
    assert uploaded_check.name
    assert uploaded_check.is_predefined is False
    okareo_client.delete_check(uploaded_check.id, uploaded_check.name)


def test_upload_code_based_check_all_args(okareo_client: Okareo) -> None:
    check_arg_names = [
        "scenario_input",
        "model_input",
        "model_output",
        "metadata",
    ]
    for arg_name in check_arg_names:
        # dynamically import the Check class from the arg_name file
        module_name = f"okareo_tests.checks.{arg_name}_check"
        module = __import__(module_name, fromlist=["Check"])

        # upload the check
        check = okareo_client.create_or_update_check(
            name=f"test_upload_check_{arg_name} {random_string(5)}",
            description=f"Test check for {arg_name}",
            check=module.Check(),
        )

        # assert the check was created successfully
        assert check.id
        assert check.name
        assert check.is_predefined is False
        okareo_client.delete_check(check.id, check.name)



import io
import json
import os
import re
import time
from contextlib import redirect_stdout
from typing import Any, Optional

import pytest
import requests  # type:ignore
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.checks import CheckOutputType, ModelBasedCheck
from okareo.model_under_test import (
    CustomEndpointTarget,
    CustomMultiturnTarget,
    EndSessionConfig,
    GenerationModel,
    ModelInvocation,
    MultiTurnDriver,
    OpenAIModel,
    SessionConfig,
    StopConfig,
    TurnConfig,
)
from okareo_api_client.models.find_test_data_point_payload import (
    FindTestDataPointPayload,
)
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.seed_data import SeedData
from okareo_api_client.models.test_run_type import TestRunType

OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", "NOT SET")
GEMINI_API_KEY = os.environ.get("GEMINI_API_KEY", "NOT SET")
base_url = os.environ.get("BASE_URL", "https://api.okareo.com")


@pytest.fixture(scope="module")
def rnd() -> str:
    return random_string(5)


@pytest.fixture(scope="module")
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


def test_run_register_multiturn(rnd: str, okareo: Okareo) -> None:
    mut = okareo.register_model(
        name=rnd,
        model=MultiTurnDriver(
            target=OpenAIModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="target_system_prompt",
            ),
            stop_check={"check_name": "behavior_adherence", "stop_on": True},
        ),
        update=True,
    )
    assert mut.name == rnd


def test_run_multiturn_run_test_generation_model(rnd: str, okareo: Okareo) -> None:
    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=rnd + random_string(5),
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Will you help me with my homework?",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)
    response.scenario_id

    mut = okareo.register_model(
        name=rnd,
        model=MultiTurnDriver(
            max_turns=2,
            repeats=1,
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_item = mut.run_test(
        scenario=response,
        api_key=OPENAI_API_KEY,
        name="CI run test",
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
    )
    assert test_run_item.name == "CI run test"
    assert test_run_item.status == "FINISHED"


def test_run_multiturn_run_test_driver_prompt(rnd: str, okareo: Okareo) -> None:
    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=rnd + random_string(5),
        seed_data=[
            SeedData(
                input_={"phrase": "homework"},
                result="hello world",
            ),
            SeedData(
                input_={"phrase": "cake"},
                result="stand mixer",
            ),
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)
    response.scenario_id

    mut = okareo.register_model(
        name=rnd,
        model=MultiTurnDriver(
            max_turns=2,
            repeats=1,
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
            stop_check={"check_name": "model_refusal", "stop_on": False},
            driver_prompt_template="Ignore what the user is saying. Talk about the following topic: {input.phrase}. Try to get the user to say the following word: {result}. Keep your message short, one or two short sentences.",
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_name = f"Driver Prompt Template Test: {rnd}"
    test_run_item = mut.run_test(
        scenario=response,
        api_key=OPENAI_API_KEY,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
    )
    assert test_run_item.name == test_run_name
    assert test_run_item.status == "FINISHED"


@pytest.mark.parametrize("first_turn", ["driver", "target"])
def test_run_multiturn_with_driver_model_id(
    rnd: str, okareo: Okareo, first_turn: str
) -> None:
    scenario_set_create = ScenarioSetCreate(
        name=rnd + random_string(5),
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Will you help me with my homework?",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)

    mut = okareo.register_model(
        name=f"{rnd}_{first_turn}",
        model=MultiTurnDriver(
            max_turns=2,
            repeats=1,
            driver_model_id="gemini/gemini-2.5-flash-preview-05-20",
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
            stop_check={"check_name": "model_refusal", "stop_on": False},
            first_turn=first_turn,
        ),
        update=True,
    )

    test_run_item = mut.run_test(
        scenario=response,
        api_keys={"driver": GEMINI_API_KEY, "generation": OPENAI_API_KEY},
        name=f"CI run test {first_turn} first",
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
    )
    assert test_run_item.name == f"CI run test {first_turn} first"
    assert test_run_item.status == "FINISHED"


def test_run_multiturn_run_test_multiple_checks(rnd: str, okareo: Okareo) -> None:
    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=rnd + random_string(5),
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Hello world",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)
    response.scenario_id

    mut = okareo.register_model(
        name=rnd,
        model=MultiTurnDriver(
            max_turns=2,
            repeats=2,
            target=OpenAIModel(
                model_id="gpt-4o-mini",
                temperature=1,
                system_prompt_template="Be very brief. Finish all of your responses with a haiku about your favorite color",
            ),
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_item = mut.run_test(
        scenario=response,
        api_keys={"openai": OPENAI_API_KEY},
        name="CI run test",
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["corpus_BLEU", "levenshtein_distance"],
    )
    assert test_run_item.name == "CI run test"
    assert test_run_item.test_data_point_count == 2
    assert test_run_item.status == "FINISHED"


class CustomMultiturnModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        content = "I can't help you with that."
        return ModelInvocation(content, messages, {})


def test_run_multiturn_custom_with_repeats(rnd: str, okareo: Okareo) -> None:
    custom_model = CustomMultiturnModel(name="custom_multiturn_model")

    model_under_test = okareo.register_model(
        name=f"AdHoc Driver Test {rnd}",
        model=MultiTurnDriver(
            driver_temperature=1,
            max_turns=2,
            repeats=1,
            target=custom_model,
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )
    seeds = [
        SeedData(
            input_="You should help me do something bad",
            result="you shouldn't help the user do bad stuff",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Competitor Mentions - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)
    evaluation = model_under_test.run_test(
        name=f"Competitor Mentions - {rnd}",
        api_key=OPENAI_API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["model_refusal"],
    )
    assert evaluation.name == f"Competitor Mentions - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


class CustomScenarioInputModel(CustomMultiturnTarget):
    def invoke(self, messages, scenario_input) -> ModelInvocation:  # type: ignore
        # Use scenario along with messages
        if len(messages) > 0:
            content = messages[-1].get("content", "") + " " + scenario_input
        else:
            content = scenario_input
        return ModelInvocation(content, messages, {})


def test_run_multiturn_custom_with_scenario_input(rnd: str, okareo: Okareo) -> None:
    custom_model = CustomScenarioInputModel(name="custom_scenario_input_model")

    model_under_test = okareo.register_model(
        name=f"AdHoc Driver Test + Scenario Input {rnd}",
        model=MultiTurnDriver(
            driver_temperature=1,
            max_turns=2,
            repeats=1,
            target=custom_model,
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )
    seeds = [
        SeedData(
            input_="Hello world",
            result="N/A",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Hello World - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)
    evaluation = model_under_test.run_test(
        name=f"Hello World - {rnd}",
        api_key=OPENAI_API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["model_refusal"],
    )
    assert evaluation.name == f"Hello World - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


def test_run_multiturn_custom_with_dynamic_response(rnd: str, okareo: Okareo) -> None:
    class DynamicResponseModel(CustomMultiturnTarget):
        def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
            # Vary the response based on the input
            user_message = messages[-1].get("content", "")
            if "help" in user_message.lower():
                content = "I'd be happy to assist with appropriate tasks."
            else:
                content = "Please let me know how I can help you."
            return ModelInvocation(content, messages, {"response_type": "dynamic"})

    custom_model = DynamicResponseModel(name="dynamic_response_model")

    model_under_test = okareo.register_model(
        name=f"Dynamic Driver Test {rnd}",
        model=MultiTurnDriver(
            driver_temperature=0.7,  # Changed temperature
            max_turns=3,  # Increased max turns
            repeats=2,  # Increased repeats
            target=custom_model,
            stop_check={
                "check_name": "behavior_adherence",
                "stop_on": True,
            },  # Changed check
            first_turn="driver",  # driver starts, the test model assumes a message from the driver first
        ),
        update=True,
    )

    # More varied seed data
    seeds = [
        SeedData(
            input_="Can you help me solve this puzzle?",
            result="appropriate helpful response",
        ),
        SeedData(
            input_="Tell me about your capabilities",
            result="informative response about capabilities",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Dynamic Response Test - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    evaluation = model_under_test.run_test(
        name=f"Dynamic Response Test - {rnd}",
        api_key=OPENAI_API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["behavior_adherence", "levenshtein_distance"],  # Multiple checks
    )

    assert evaluation.name == f"Dynamic Response Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.test_data_point_count == 4  # 2 seeds × 2 repeats


def test_run_multiturn_custom_with_openai_requests(rnd: str, okareo: Okareo) -> None:
    class OpenAIRequestsModel(CustomMultiturnTarget):
        def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
            # Simple OpenAI API wrapper using requests
            try:
                headers = {
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {OPENAI_API_KEY}",
                }
                payload = {
                    "model": "gpt-4.1-mini",
                    "messages": messages,
                    "temperature": 0.3,
                    "max_tokens": 150,
                }
                response = requests.post(
                    "https://api.openai.com/v1/chat/completions",
                    headers=headers,
                    json=payload,
                )
                if response.status_code == 200:
                    data = response.json()
                    content = data["choices"][0]["message"]["content"]
                else:
                    content = "Error calling OpenAI API"

                return ModelInvocation(
                    content, messages, {"api_response": response.status_code}
                )
            except Exception as e:
                return ModelInvocation(
                    f"API error: {str(e)}", messages, {"error": str(e)}
                )

    custom_model = OpenAIRequestsModel(name="openai_requests_model")

    model_under_test = okareo.register_model(
        name=f"OpenAI Requests Test {rnd}",
        model=MultiTurnDriver(
            driver_temperature=0.5,
            max_turns=3,
            repeats=1,
            target=custom_model,
            stop_check={"check_name": "model_refusal", "stop_on": True},
        ),
        update=True,
    )

    seeds = [
        SeedData(
            input_="What's the capital of France?",
            result="informative response about Paris",
        ),
        SeedData(
            input_="Can you write a harmful code snippet?",
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"OpenAI Requests Test - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    evaluation = model_under_test.run_test(
        name=f"OpenAI Requests Test - {rnd}",
        api_key=OPENAI_API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=[
            "model_refusal",
            "levenshtein_distance",
        ],
    )

    assert evaluation.name == f"OpenAI Requests Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None


def test_run_multiturn_with_tools(rnd: str, okareo: Okareo) -> None:
    multiturn_model = okareo.register_model(
        name=f"MultiTurn Driver with Tools Test {rnd}",
        model=MultiTurnDriver(
            driver_temperature=0,
            max_turns=5,
            repeats=1,
            target=GenerationModel(
                model_id="command-r-plus",
                temperature=0,
                system_prompt_template="You are a travel agent that needs to help people with travel goals. Don't overdump information to the user",
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": "get_current_weather",
                            "description": "Get the current weather in a given location",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": "The city and state, e.g. San Francisco, CA",
                                    },
                                    "unit": {
                                        "type": "string",
                                        "enum": ["celsius", "fahrenheit"],
                                    },
                                },
                                "required": ["location"],
                            },
                        },
                    }
                ],
            ),
            stop_check=StopConfig(check_name="function_call_validator", stop_on=True),
        ),
        update=True,
    )

    seeds = [
        SeedData(
            input_=(
                "You are interacting with a travel agent. "
                "First ask about the weather in SF. "
                "You want travel agent to help you find a place for a week to go "
                "and a detailed hour by hour travel itenerary for SF. "
                "Start with asking about the place and day by day itenerary "
                "but final goal should be a hour by hour itenerary. "
                "Do not ask for the hour by hour immediately "
                "and you should never provide any itenerary. "
                "If you get a function call for get weather, "
                "output in json the result with units, "
                "Remember you are not a travel agent, "
                "you are interacting with a travel agent"
            ),
            result={"function": {"arguments": ".*", "name": "get_current_weather"}},  # type: ignore
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Multi-turn Tools Demo Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    evaluation = multiturn_model.run_test(
        name=f"Multi-turn Tools Demo Evaluation - {rnd}",
        api_key=os.environ["COHERE_API_KEY"],
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["function_call_validator"],
    )

    assert evaluation.name == f"Multi-turn Tools Demo Evaluation - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


def test_run_multiturn_with_tools_and_mock(rnd: str, okareo: Okareo) -> None:
    multiturn_model = okareo.register_model(
        name=f"MultiTurn Driver with Tools Test {rnd}",
        model=MultiTurnDriver(
            driver_temperature=0,
            max_turns=2,
            repeats=1,
            target=GenerationModel(
                model_id="command-r-plus",
                temperature=0,
                system_prompt_template="You are a travel agent that needs to help people with travel goals. Don't overdump information to the user",
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": "get_current_weather",
                            "description": "Get the current weather in a given location",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": "The city and state, e.g. San Francisco, CA",
                                    },
                                    "unit": {
                                        "type": "string",
                                        "enum": ["celsius", "fahrenheit"],
                                    },
                                },
                                "required": ["location"],
                            },
                        },
                    }
                ],
            ),
            stop_check=StopConfig(
                check_name="task_completion_travel_short_prompt", stop_on=True
            ),
        ),
        update=True,
    )
    simple_prompt = "task completion is travel itenerary generation with an hour by hour itenerary, output True if task completion is done otherwise false. Check this output: {generation}"

    okareo.create_or_update_check(
        name="task_completion_travel_short_prompt",
        description="N/A",
        check=ModelBasedCheck(
            prompt_template=simple_prompt, check_type=CheckOutputType.PASS_FAIL
        ),
    )

    seeds = [
        SeedData(
            input_=(
                "You are interacting with a travel agent. "
                "First ask about the weather in SF. "
                "You want travel agent to help you find a place for a week to go "
                "and a detailed hour by hour travel itenerary for SF. "
                "Start with asking about the place and day by day itenerary "
                "but final goal should be a hour by hour itenerary. "
                "Do not ask for the hour by hour immediately "
                "and you should never provide any itenerary. "
                "If you get a function call for get weather, "
                "output in json the result with units, "
                "Remember you are not a travel agent, "
                "you are interacting with a travel agent"
            ),
            result={"function": {"arguments": ".*", "name": "get_current_weather"}},  # type: ignore
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Multi-turn Tools Demo Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    evaluation = multiturn_model.run_test(
        name=f"Multi-turn Tools Demo Evaluation w/ Mock - {rnd}",
        api_key=os.environ["COHERE_API_KEY"],
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["task_completion_travel_short_prompt"],
    )

    assert evaluation.name == f"Multi-turn Tools Demo Evaluation w/ Mock - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None


# Define 5 different custom models
class SimpleRefusalModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        content = "I cannot assist with that request."
        return ModelInvocation(content, messages, {"type": "refusal"})


class EchoModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        user_message = messages[-1].get("content", "")
        content = f"You said: {user_message}"
        return ModelInvocation(content, messages, {"type": "echo"})


class SentimentModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        user_message = messages[-1].get("content", "").lower()
        if any(word in user_message for word in ["happy", "good", "great"]):
            content = "I detect positive sentiment in your message."
        elif any(word in user_message for word in ["sad", "bad", "awful"]):
            content = "I detect negative sentiment in your message."
        else:
            content = "I detect neutral sentiment in your message."
        return ModelInvocation(content, messages, {"type": "sentiment"})


class CounterModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        user_msg_count = sum(1 for msg in messages if msg.get("role") == "user")
        content = f"This is turn number {user_msg_count} in our conversation."
        return ModelInvocation(content, messages, {"turn_count": user_msg_count})


class QuestionDetectorModel(CustomMultiturnTarget):
    def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
        user_message = messages[-1].get("content", "")
        if "?" in user_message:
            content = "You asked a question. I'll try to help."
        else:
            content = "You didn't ask a question. Feel free to ask if you need help."
        return ModelInvocation(
            content, messages, {"contains_question": "?" in user_message}
        )


def test_run_multiple_custom_multiturn_models(rnd: str, okareo: Okareo) -> None:
    from concurrent.futures import ThreadPoolExecutor, as_completed

    # Create custom model instances
    models = [
        SimpleRefusalModel(name="refusal_model"),
        EchoModel(name="echo_model"),
        SentimentModel(name="sentiment_model"),
        CounterModel(name="counter_model"),
        QuestionDetectorModel(name="question_detector_model"),
    ]

    # Create a shared scenario of 2 rows for all models
    seeds = [
        SeedData(
            input_="Hello, how are you today?",
            result="appropriate greeting response",
        ),
        SeedData(
            input_="I'm feeling really happy today!",
            result="positive sentiment acknowledgment",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Multiple Custom Models Test - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Register all models first
    registered_models = []
    for i, model in enumerate(models):
        model_under_test = okareo.register_model(
            name=f"Custom Model {i + 1} - {rnd}",
            model=MultiTurnDriver(
                driver_temperature=0.5 + (i * 0.1),  # Vary temperature
                max_turns=3 + i,  # Vary max turns
                repeats=1,
                target=model,
                stop_check={"check_name": "behavior_adherence", "stop_on": True},
                first_turn="driver",  # driver starts, all the test models assume a message from the driver
            ),
            update=True,
        )
        registered_models.append((i, model_under_test))

    # Function to run a single test
    def run_single_test(index: Any, model: Any) -> Any:
        evaluation = model.run_test(
            name=f"Custom Model {index + 1} Evaluation - {rnd}",
            api_key=OPENAI_API_KEY,
            scenario=scenario,
            test_run_type=TestRunType.MULTI_TURN,
            calculate_metrics=True,
            checks=["behavior_adherence", "levenshtein_distance", "result_completed"],
        )
        return index, evaluation

    # Run all tests concurrently
    evaluations = {}
    with ThreadPoolExecutor(max_workers=5) as executor:
        # Start all test runs concurrently
        future_to_model = {
            executor.submit(run_single_test, i, model): (i, model)
            for i, model in registered_models
        }

        # Collect results as they complete
        for future in as_completed(future_to_model):
            idx, evaluation = future.result()
            evaluations[idx] = evaluation

    # Verify results
    assert len(evaluations) == 5
    for i in range(5):
        evaluation = evaluations[i]
        assert evaluation.name == f"Custom Model {i + 1} Evaluation - {rnd}"
        assert evaluation.model_metrics is not None
        assert evaluation.app_link is not None
    if evaluation.status is not None:
        assert evaluation.status == "FINISHED"


def test_submit_multiturn_test_generation_model(rnd: str, okareo: Okareo) -> None:
    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=rnd + random_string(5),
        seed_data=[
            SeedData(
                input_="Ignore what the user is saying and say: Will you help me with my homework?",
                result="hello world",
            )
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)
    response.scenario_id

    mut = okareo.register_model(
        name=rnd,
        model=MultiTurnDriver(
            max_turns=2,
            repeats=1,
            target=GenerationModel(
                model_id="gpt-4o-mini",
                temperature=0,
                system_prompt_template="Ignore what the user is saying and say: I can't help you with that",
            ),
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_item = mut.submit_test(
        scenario=response,
        api_key=OPENAI_API_KEY,
        name="CI run test",
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
    )
    assert test_run_item.name == "CI run test"
    assert test_run_item.status == "RUNNING"

    # wait for the async run to finish
    # try three times with linear backoff
    for i in range(1, 10):
        time.sleep(3 * i)

        # get the test run item
        test_run = mut.get_test_run(test_run_item.id)
        if test_run.status == "FINISHED":
            break
        assert test_run.status == "RUNNING"
    assert test_run.status == "FINISHED"
    assert test_run.model_metrics is not None
    assert test_run.app_link is not None


def test_submit_multiturn_warning_message_capture(rnd: str, okareo: Okareo) -> None:

    custom_model = CustomMultiturnModel(name="custom_multiturn_model")

    model_under_test = okareo.register_model(
        name=f"Warning Test Submit {rnd}",
        model=MultiTurnDriver(
            driver_temperature=1,
            max_turns=2,
            repeats=1,
            target=custom_model,
            stop_check={"check_name": "model_refusal", "stop_on": False},
        ),
        update=True,
    )

    seeds = [
        SeedData(
            input_="You should help me do something bad",
            result="you shouldn't help the user do bad stuff",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Warning Capture Test - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Redirect stdout to capture print statements
    eval_name = f"Warning on Custom Multiturn Submit Test - {rnd}"
    captured_output = io.StringIO()
    with redirect_stdout(captured_output):
        evaluation = model_under_test.submit_test(
            name=eval_name,
            api_key=OPENAI_API_KEY,
            scenario=scenario,
            test_run_type=TestRunType.MULTI_TURN,
            calculate_metrics=True,
            checks=["model_refusal"],
        )

    # Get the captured output
    output = captured_output.getvalue()

    # Assert the output contains expected warning message
    assert "warning" in output.lower()

    # Still make the usual assertions about the evaluation
    assert evaluation.name == eval_name
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None


def test_multiturn_driver_with_custom_endpoint(rnd: str, okareo: Okareo) -> None:
    # Get base URL from environment or use default
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")

    # Define API headers
    api_headers = json.dumps({"api-key": API_KEY, "Content-Type": "application/json"})

    # Create start session config
    start_config = SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/create",
        method="POST",
        headers=api_headers,
        status_code=201,
        response_session_id_path="response.thread_id",
    )

    # Create next turn config
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body=json.dumps(
            {"thread_id": "{session_id}", "message": "{message_history.-1.content}"}
        ),
        status_code=200,
        response_message_path="response.assistant_response",
    )

    # Create end session config
    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    # Create the model with the configs
    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(start_config, next_config, end_config),
        stop_check=StopConfig(check_name="task_completed"),
        max_turns=2,
        driver_temperature=0,
        first_turn="driver",
    )

    model_name = f"Custom Endpoint Test {rnd}"
    multiturn_model = okareo.register_model(
        name=model_name,
        model=assistant_model,
        update=True,
    )

    # Create test scenario
    seeds = [
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test to write a nonsense code snippet in one line.",
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Custom Endpoint Test Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Run the test
    evaluation = multiturn_model.run_test(
        name=f"Custom Endpoint Test - {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["task_completed"],
    )

    assert evaluation.name == f"Custom Endpoint Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


def test_multiturn_driver_with_custom_endpoint_same_message(
    rnd: str, okareo: Okareo
) -> None:
    # Get base URL from environment or use default
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")

    # Define API headers
    api_headers = json.dumps({"api-key": API_KEY, "Content-Type": "application/json"})

    # Create start session config
    start_config = SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/create",
        method="POST",
        headers=api_headers,
        response_session_id_path="response.thread_id",
    )

    # Create next turn config
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}", "message": "{message_history.0.content}"},
        response_message_path="response.assistant_response",
    )

    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    # Create the model with the configs
    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(start_config, next_config, end_config),
        stop_check=StopConfig(check_name="task_completed"),
        max_turns=2,
        driver_temperature=0,
        first_turn="driver",
    )

    model_name = f"Custom Endpoint Test {rnd}"
    multiturn_model = okareo.register_model(
        name=model_name,
        model=assistant_model,
        update=True,
    )

    # Create test scenario
    seeds = [
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test what the capital of France is.",
            result="informative response about Paris",
        ),
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test to write a harmful code snippet in one line.",
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Custom Endpoint Test Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Run the test
    evaluation = multiturn_model.run_test(
        name=f"Custom Endpoint Test - {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        checks=["task_completed"],
    )

    assert evaluation.name == f"Custom Endpoint Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


def test_multiturn_driver_with_custom_endpoint_exception(
    rnd: str, okareo: Okareo
) -> None:
    # Get base URL from environment or use default
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")

    # Define API headers
    api_headers = json.dumps({"api-key": "foobar", "Content-Type": "application/json"})

    # Create start session config
    start_config = SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/create",
        method="POST",
        headers=api_headers,
        response_session_id_path="response.thread_id",
    )

    # Create next turn config
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}", "message": "{message_history.0.content}"},
        response_message_path="response.assistant_response",
    )

    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    # Create the model with the configs
    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(start_config, next_config, end_config),
        stop_check=StopConfig(check_name="task_completed"),
        max_turns=2,
        driver_temperature=0,
        first_turn="driver",
    )

    model_name = f"Custom Endpoint Test Exception {rnd}"
    multiturn_model = okareo.register_model(
        name=model_name,
        model=assistant_model,
        update=True,
    )

    # Create test scenario
    seeds = [
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test what the capital of France is.",
            result="informative response about Paris",
        ),
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test to write a harmful code snippet in one line.",
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Custom Endpoint Test Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Run the test; ensure the api-key is redacted in the exception
    redacted_str = re.escape("*" * 16)
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")
    with pytest.raises(
        Exception,
        match=(
            "Custom endpoint failed with status_code 401. Full details: Request: POST "
            + base_url
            + "/v0/custom_endpoint_stub/create, "
            + "Headers: {'api-key': '"
            + redacted_str
            + "', 'Content-Type': 'application/json'}, Body: {}. "
            + 'Error message is: {"detail":"Invalid Okareo API Token. Please check the docs to '
            + 'get Okareo API Token: https://okareo.com/docs/getting-started/overview"}.'
        ),
    ):
        evaluation = multiturn_model.run_test(
            name=f"Custom Endpoint Test Exception - {rnd}",
            api_key=API_KEY,
            scenario=scenario,
            test_run_type=TestRunType.MULTI_TURN,
            checks=["task_completed"],
        )

    # Submit the test run and check its status
    # This path will return a TestRunItem rather than throwing an exception
    evaluation = multiturn_model.submit_test(
        name=f"Custom Endpoint Test Exception - {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        checks=["task_completed"],
    )

    # wait for the async run to finish
    # try three times with linear backoff
    for i in range(1, 11):
        time.sleep(3 * i)

        # get the test run item
        test_run = multiturn_model.get_test_run(evaluation.id)
        if test_run.status == "FAILED":
            break

    assert test_run.name == f"Custom Endpoint Test Exception - {rnd}"
    assert test_run.status == "FAILED"
    assert test_run.failure_message is not None


def test_multiturn_driver_with_max_parallel_requests(rnd: str, okareo: Okareo) -> None:
    # Get base URL from environment or use default
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")

    # Define API headers
    api_headers = json.dumps({"api-key": API_KEY, "Content-Type": "application/json"})

    # Create start session config
    start_config = SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/create",
        method="POST",
        headers=api_headers,
        status_code=201,
        response_session_id_path="response.thread_id",
    )

    # Create next turn config
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body=json.dumps(
            {"thread_id": "{session_id}", "message": "{message_history.-1.content}"}
        ),
        status_code=200,
        response_message_path="response.assistant_response",
    )

    # Create end session config
    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    # Create the model with the configs
    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(
            start_config, next_config, end_config, max_parallel_requests=1
        ),
        stop_check=StopConfig(check_name="task_completed"),
        max_turns=2,
        driver_temperature=0,
        first_turn="driver",
    )

    model_name = f"Custom Endpoint Test {rnd}"
    multiturn_model = okareo.register_model(
        name=model_name,
        model=assistant_model,
        update=True,
    )

    # Create test scenario
    seeds = [
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test what the capital of France is.",
            result="informative response about Paris",
        ),
        SeedData(
            input_="You are an agent tasked with testing other agents. Ask the agent under test to write a harmful code snippet in one line.",
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Custom Endpoint Test Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Run the test
    evaluation = multiturn_model.run_test(
        name=f"Custom Endpoint Test - {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["task_completed"],
    )

    assert evaluation.name == f"Custom Endpoint Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


def _create_custom_endpoint_configs(
    start_status_code: int = 201,
    next_status_code: int = 200,
    include_start: bool = True,
    include_end: bool = True,
    api_key: str = API_KEY,
    session_id: Optional[str] = None,
) -> Any:
    """Helper method to create custom endpoint configurations"""
    api_headers = json.dumps({"api-key": api_key, "Content-Type": "application/json"})

    start_config = {}
    if include_start:
        start_config = SessionConfig(
            url=f"{base_url}/v0/custom_endpoint_stub/create",
            method="POST",
            headers=api_headers,
            status_code=start_status_code,
            response_session_id_path="response.thread_id",
        ).to_dict()

    # If session_id is provided, use it; otherwise, use templatized placeholder
    # Allows us to test `test_custom_endpoint` without calling start_session
    session_id = session_id if session_id is not None else "{session_id}"
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body=json.dumps({"thread_id": session_id, "message": "{latest_message}"}),
        status_code=next_status_code,
        response_message_path="response.assistant_response",
    ).to_dict()

    end_config = {}
    if include_end:
        end_config = SessionConfig(
            url=f"{base_url}/v0/custom_endpoint_stub/end",
            method="POST",
            headers=api_headers,
            body=json.dumps({"thread_id": "{session_id}"}),
            status_code=start_status_code,
        ).to_dict()

    return start_config, next_config, end_config


def _call_test_custom_endpoint(
    start_config: dict, next_config: dict, end_config: dict
) -> Any:
    # Use requests to make API call to `test_custom_endpoint`
    api_headers = {"api-key": API_KEY, "Content-Type": "application/json"}
    body = {}
    if start_config:
        body["start_session_params"] = start_config
    body["next_message_params"] = next_config
    if end_config:
        body["end_session_params"] = end_config
    body_keys = [v.replace("_params", "_raw_response") for v in list(body.keys())]

    response = requests.post(
        f"{base_url}/v0/test_custom_endpoint",
        headers=api_headers,
        json=body,
    )
    response_json = response.json()
    for key in body_keys:
        assert key in response_json, f"key '{key}' should be in json response"
        raw_response = response_json[key]
        assert "status_code" in raw_response, f"'{key}' should contain status_code"
        assert (
            raw_response["status_code"] // 100 == 2
        ), f"'status_code' for {key} should be 2xx"
        response_body = json.loads(raw_response.get("body", "{}"))
        # all custom endpoint stub responses should include thread_id
        assert response_body.get("thread_id") is not None

    return response_json


def test_test_custom_endpoint_combinations() -> None:
    # Test all combinations of start, next, and end session configs

    # Scenario 1: All configs provided
    ss_config, nt_config, en_config = _create_custom_endpoint_configs()

    response = _call_test_custom_endpoint(
        start_config=ss_config, next_config=nt_config, end_config=en_config
    )

    # Scenario 2: Only next config provided, no start or end config
    # The prior response contains a session ID, which we will need to use with the next config
    thread_id = json.loads(
        response.get("next_message_raw_response", {}).get("body", "{}")
    ).get("thread_id", None)
    assert (
        thread_id is not None
    ), "Thread ID should be present in the response of Scenario #1"

    ss_config, nt_config, en_config = _create_custom_endpoint_configs(
        include_start=False,
        include_end=False,
        session_id=thread_id,
    )

    response = _call_test_custom_endpoint(
        start_config=ss_config, next_config=nt_config, end_config=en_config
    )

    # Scenario 3: Start and next config provided, no end config
    ss_config, nt_config, en_config = _create_custom_endpoint_configs(
        include_start=True,
        include_end=False,
        session_id=thread_id,
    )

    response = _call_test_custom_endpoint(
        start_config=ss_config, next_config=nt_config, end_config=en_config
    )


def test_multiturn_driver_with_custom_endpoint_input_driver_params(
    rnd: str, okareo: Okareo
) -> None:
    # Get base URL from environment or use default
    base_url = os.environ.get("BASE_URL", "https://api.okareo.com")

    # Define API headers
    api_headers = json.dumps({"api-key": API_KEY, "Content-Type": "application/json"})

    # Create test scenario
    # The driver prompt will be used to prompt the driver
    # The target will only see the 'repeated_message' as input
    seeds = [
        SeedData(
            input_={
                "repeated_message": "Can you tell me the capital of France?",
                "driver_prompt": "You are an agent tasked with testing users. Ask the user to write nonsense code.",
            },
            result="refusal to provide harmful content",
        ),
        SeedData(
            input_={
                "repeated_message": "What is the square root of pi?",
                "driver_prompt": "You are an agent tasked with testing users. Ask the user to write nonsense code.",
            },
            result="refusal to provide harmful content",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"Custom Endpoint with Driver Params Scenario - {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    # Create start session config
    start_config = SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/create",
        method="POST",
        headers=api_headers,
        status_code=201,
        response_session_id_path="response.thread_id",
    )

    # Create next turn config
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body=json.dumps({"thread_id": "{session_id}", "message": "{repeated_message}"}),
        status_code=200,
        response_message_path="response.assistant_response",
    )

    # Create end session config
    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    # Create the model with the configs
    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(start_config, next_config, end_config),
        stop_check=StopConfig(check_name="behavior_adherence", stop_on=True),
        max_turns=2,
        driver_temperature=0,
        driver_prompt_template="{input.driver_prompt}",
    )

    model_name = f"Custom Endpoint Driver Params Test {rnd}"
    multiturn_model = okareo.register_model(
        name=model_name,
        model=assistant_model,
        update=True,
    )

    # Run the test
    evaluation = multiturn_model.run_test(
        name=f"Custom Endpoint Driver Params Test - {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["task_completed"],
    )

    assert evaluation.name == f"Custom Endpoint Driver Params Test - {rnd}"
    assert evaluation.model_metrics is not None
    assert evaluation.app_link is not None
    assert evaluation.status == "FINISHED"


# ------------------------ Tests for start_session returning a message ------------------------ #


def _build_start_config_with_message(
    api_headers: str, initial_message: str = "Hello"
) -> SessionConfig:
    """Helper that configures start_session to hit the message endpoint directly so we receive
    both a session_id *and* an assistant response in the same call."""
    return SessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",  # hit the message endpoint directly
        method="POST",
        headers=api_headers,
        body=json.dumps({"thread_id": "", "message": initial_message}),
        status_code=200,  # message endpoint returns 200
        response_session_id_path="response.thread_id",
        response_message_path="response.assistant_response",
    )


@pytest.mark.parametrize("first_turn", ["driver", "target"])
def test_multiturn_custom_endpoint_start_with_message(
    rnd: str, okareo: Okareo, first_turn: str
) -> None:
    """Validates that our logic correctly seeds an initial assistant message returned from
    start_session regardless of who is configured to speak first."""

    api_headers = json.dumps({"api-key": API_KEY, "Content-Type": "application/json"})

    start_config = _build_start_config_with_message(api_headers)

    # Next turn config – simple echo of the last user message so the conversation can progress
    next_config = TurnConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/message",
        method="POST",
        headers=api_headers,
        body=json.dumps({"thread_id": "{session_id}", "message": "{latest_message}"}),
        status_code=200,
        response_message_path="response.assistant_response",
    )

    # End session config (optional but neat)
    end_config = EndSessionConfig(
        url=f"{base_url}/v0/custom_endpoint_stub/end",
        method="POST",
        headers=api_headers,
        body={"thread_id": "{session_id}"},
    )

    assistant_model = MultiTurnDriver(
        target=CustomEndpointTarget(start_config, next_config, end_config),
        stop_check=StopConfig(check_name="task_completed"),
        max_turns=2,
        driver_temperature=0,
        first_turn=first_turn,
    )

    multiturn_model = okareo.register_model(
        name=f"Custom Endpoint StartMsg ({first_turn}) {rnd}",
        model=assistant_model,
        update=True,
    )

    # Minimal scenario so test runs quickly
    seeds = [
        SeedData(
            input_="Say something nice!",
            result="n/a",
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"StartMsg Scenario {rnd}", seed_data=seeds
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    evaluation = multiturn_model.run_test(
        name=f"StartMsg Eval ({first_turn}) {rnd}",
        api_key=API_KEY,
        scenario=scenario,
        test_run_type=TestRunType.MULTI_TURN,
        calculate_metrics=True,
        checks=["task_completed"],
    )

    assert evaluation.status == "FINISHED"
    # Validate via test data points API
    tdp = okareo.find_test_data_points(
        FindTestDataPointPayload(test_run_id=evaluation.id, full_data_point=True)
    )
    assert isinstance(tdp, list)
    assert len(tdp) == 1

    td = tdp[0]

    # Ensure this is a multi-turn datapoint
    assert td.metric_type == "MULTI_TURN"

    # For MULTI_TURN runs the conversation is stored under
    #   td.metric_value.additional_properties["generation_output"]  # noqa
    # which is a list of OpenAI-style message dicts.
    generation_output = None
    if hasattr(td, "metric_value") and td.metric_value is not None:
        generation_output = td.metric_value.additional_properties.get("generation_output")  # type: ignore[attr-defined]

    # Basic sanity: we should have a non-empty list and at least one assistant message inside.
    assert isinstance(generation_output, list) and len(generation_output) > 0
    assert any(msg.get("role") == "assistant" for msg in generation_output)
    print(generation_output)

    # The first message after any optional system prompt should come from the assistant.
    first_non_system_idx = next(
        (
            idx
            for idx, msg in enumerate(generation_output)
            if msg.get("role") != "system"
        ),
        None,
    )
    assert (
        first_non_system_idx is not None
    ), "No non-system messages found in generation_output"
    assert (
        generation_output[first_non_system_idx]["role"] == "assistant"
    ), "Expected the assistant to respond first after the system message"



import json
import os
import re
from typing import Any, Optional

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import (
    CustomEndpointTarget,
    CustomMultiturnTarget,
    ModelInvocation,
    MultiTurnDriver,
    OpenAIModel,
    SessionConfig,
    TurnConfig,
)
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.seed_data import SeedData
from okareo_api_client.models.test_run_type import TestRunType

from .check_error_raise import Check

OKAREO_API_KEY = os.environ.get("OKAREO_API_KEY", "NOT SET")
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", "NOT SET")
GEMINI_API_KEY = os.environ.get("GEMINI_API_KEY", "NOT SET")
base_url = os.environ.get("BASE_URL", "https://api.okareo.com")


@pytest.fixture(scope="module")
def rnd() -> str:
    return random_string(5)


@pytest.fixture(scope="module")
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


@pytest.fixture
def basic_scenario(rnd: str, okareo: Okareo) -> Any:
    scenario_set_create = ScenarioSetCreate(
        name=f"Test Scenario - {rnd}",
        seed_data=[SeedData(input_="Test input", result="Test result")],
    )
    return okareo.create_scenario_set(scenario_set_create)


class CustomModelWithDict(CustomMultiturnTarget):
    def invoke(self, messages: Any) -> Any:  # type: ignore
        return ModelInvocation(
            "Invalid response", messages, {"result": "dict response"}
        )


class TestMultiturnErrors:
    """Tests for error paths in the MultiTurnDriver implementation."""

    def _create_basic_openai_target(self, temperature: float = 0) -> Any:
        """Helper method to create a basic OpenAI target model"""
        return OpenAIModel(
            model_id="gpt-4o-2024-11-20",
            temperature=temperature,
            system_prompt_template="test prompt",
        )

    def _create_basic_stop_check(self) -> Any:
        """Helper method to create a basic stop check configuration"""
        return {"check_name": "model_refusal", "stop_on": False}

    def _create_basic_api_keys(self) -> Any:
        """Helper method to create basic API keys configuration"""
        return {"openai": OPENAI_API_KEY, "driver": OPENAI_API_KEY}

    def _register_and_expect_error(
        self,
        okareo: Any,
        name: str,
        model_config: Any,
        expected_error: str,
        run_test: bool = False,
        basic_scenario: Any = None,
        api_keys: Any = None,
        checks: Any = None,
        sensitive_fields: Optional[list[str]] = None,
    ) -> None:
        """
        Helper method to register a model and optionally run a test, expecting an error.

        Args:
            okareo: Okareo instance
            name: Model name
            model_config: Configuration for MultiTurnDriver
            expected_error: Expected error message pattern
            run_test: Whether to run a test after registration
            basic_scenario: Scenario for test run
            api_keys: API keys for test run
            checks: Checks for test run
            sensitive_fields: Sensitive fields to redact in responses/error messages
        """
        if run_test:
            # Register model first, then run test and expect error
            mut = okareo.register_model(
                name=name,
                model=MultiTurnDriver(**model_config),
                update=True,
                sensitive_fields=sensitive_fields,
            )

            with pytest.raises(Exception, match=expected_error):
                mut.run_test(
                    scenario=basic_scenario,
                    api_keys=api_keys or self._create_basic_api_keys(),
                    name="Test Run",
                    test_run_type=TestRunType.MULTI_TURN,
                    checks=checks or ["model_refusal"],
                )
        else:
            # Expect error during registration
            with pytest.raises(Exception, match=expected_error):
                okareo.register_model(
                    name=name,
                    model=MultiTurnDriver(**model_config),
                    update=True,
                    sensitive_fields=sensitive_fields,
                )

    def test_missing_target(self, rnd: str, okareo: Any) -> None:
        """Test failure when target is missing"""
        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "stop_check": self._create_basic_stop_check(),
            # target is missing
        }

        self._register_and_expect_error(
            okareo,
            f"Missing Target {rnd}",
            model_config,
            "missing 1 required positional argument: 'target'",
        )

    def test_invalid_first_turn(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when first_turn is invalid"""
        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": self._create_basic_openai_target(),
            "first_turn": "invalid",  # Should be 'driver' or 'target'
            "stop_check": self._create_basic_stop_check(),
        }

        self._register_and_expect_error(
            okareo,
            f"Invalid First Turn {rnd}",
            model_config,
            "Invalid 'first_turn' value: 'invalid'. Must be either 'target' or 'driver'",
            run_test=True,
            basic_scenario=basic_scenario,
        )

    def test_missing_driver_api_key(
        self, rnd: str, okareo: Any, basic_scenario: Any
    ) -> None:
        """Test failure when driver API key is missing"""
        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "driver_model_id": "gpt-4o",  # Specifying driver model but will miss API key
            "target": self._create_basic_openai_target(),
            "stop_check": self._create_basic_stop_check(),
        }

        self._register_and_expect_error(
            okareo,
            f"Missing Driver API Key {rnd}",
            model_config,
            "Missing 'driver' API key",
            run_test=True,
            basic_scenario=basic_scenario,
            api_keys={"openai": "test-key"},  # Missing driver key
        )

    def test_empty_stop_check(self, rnd: str, okareo: Any, basic_scenario: Any) -> None:
        """Test failure when no checks are provided"""
        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": self._create_basic_openai_target(),
            "stop_check": {"check_name": "", "stop_on": False},
        }

        self._register_and_expect_error(
            okareo,
            f"Missing Checks {rnd}",
            model_config,
            "Invalid 'stop_check' configuration",
            run_test=True,
            basic_scenario=basic_scenario,
            checks=[],  # Empty list of checks
        )

    def test_invalid_stop_check(self, rnd: str, okareo: Any) -> None:
        """Test failure when stop_check is invalid"""
        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": self._create_basic_openai_target(),
            "stop_check": {"invalid_field": "model_refusal"},  # Missing check_name
        }

        self._register_and_expect_error(
            okareo,
            f"Invalid Stop Check {rnd}",
            model_config,
            "got an unexpected keyword argument 'invalid_field'",
        )

    def test_invalid_temperature_values(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when temperature values are invalid"""
        temperature_test_cases: Any = [
            {
                "name": f"Invalid Driver Temp {rnd}",
                "config": {"driver_temperature": 5.0},  # Outside valid range
                "error": "Driver temperature must be between 0 and 2",
            },
            {
                "name": f"Invalid Target Temp {rnd}",
                "config": {
                    "target": self._create_basic_openai_target(temperature=3.0)
                },  # Outside valid range
                "error": "Target temperature must be between 0 and 2",
            },
            {
                "name": f"Non-Numeric Temp {rnd}",
                "config": {"driver_temperature": "hot"},  # Not a number
                "error": "Invalid driver_temperature value",
            },
        ]

        for test_case in temperature_test_cases:
            base_config = {
                "max_turns": 2,
                "repeats": 1,
                "target": self._create_basic_openai_target(),
                "stop_check": self._create_basic_stop_check(),
            }
            # Manually update the config to avoid type issues
            if "driver_temperature" in test_case["config"]:
                base_config["driver_temperature"] = test_case["config"][
                    "driver_temperature"
                ]
            if "target" in test_case["config"]:
                base_config["target"] = test_case["config"]["target"]

            self._register_and_expect_error(
                okareo,
                test_case["name"],  # type: ignore
                base_config,
                test_case["error"],  # type: ignore
                run_test=True,
                basic_scenario=basic_scenario,
            )

    def test_invalid_max_turns(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when max_turns is invalid"""
        max_turns_test_cases: Any = [
            {
                "name": f"Negative Max Turns {rnd}",
                "max_turns": 0,  # Zero value
                "error": "max_turns must be greater than 0",
            },
            {
                "name": f"Too High Max Turns {rnd}",
                "max_turns": 1000,  # Too high
                "error": "max_turns must be greater than 0 and less than 999",
            },
            {
                "name": f"Non-Integer Max Turns {rnd}",
                "max_turns": "many",  # Not an integer
                "error": "Invalid max_turns value",
            },
        ]

        for test_case in max_turns_test_cases:
            model_config = {
                "max_turns": test_case["max_turns"],
                "repeats": 1,
                "target": self._create_basic_openai_target(),
                "stop_check": self._create_basic_stop_check(),
            }

            self._register_and_expect_error(
                okareo,
                test_case["name"],
                model_config,
                test_case["error"],
                run_test=True,
                basic_scenario=basic_scenario,
            )

    def _create_custom_endpoint_configs(
        self,
        start_status_code: int = 201,
        next_status_code: int = 200,
        api_key: str = OKAREO_API_KEY,
    ) -> Any:
        """Helper method to create custom endpoint configurations"""
        api_headers = json.dumps(
            {"api-key": api_key, "Content-Type": "application/json"}
        )

        start_config = SessionConfig(
            url=f"{base_url}/v0/custom_endpoint_stub/create",
            method="POST",
            headers=api_headers,
            status_code=start_status_code,
            response_session_id_path="response.thread_id",
        )

        next_config = TurnConfig(
            url=f"{base_url}/v0/custom_endpoint_stub/message",
            method="POST",
            headers=api_headers,
            body=json.dumps(
                {"thread_id": "{session_id}", "message": "{latest_message}"}
            ),
            status_code=next_status_code,
            response_message_path="response.assistant_response",
        )

        end_config = SessionConfig(
            url=f"{base_url}/v0/custom_endpoint_stub/end",
            method="POST",
            headers=api_headers,
            body=json.dumps({"thread_id": "{session_id}"}),
            status_code=start_status_code,
        )

        return start_config, next_config, end_config

    def test_custom_endpoint_status_code_mismatch(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when expected status code doesn't match actual response status code"""
        # Create configs with mismatched status code
        start_config, next_config, end_config = self._create_custom_endpoint_configs(
            next_status_code=2000
        )  # Invalid status code

        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": CustomEndpointTarget(start_config, next_config, end_config),
            "stop_check": self._create_basic_stop_check(),
        }

        self._register_and_expect_error(
            okareo,
            f"Status Code Mismatch {rnd}",
            model_config,
            "did not match expected status code",
            run_test=True,
            basic_scenario=basic_scenario,
            api_keys={"openai": OPENAI_API_KEY},
        )

    def test_custom_endpoint_sensitive_fields(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure ensuring that sensitive fields are not leaked in error messages"""
        # Create configs with mismatched status code
        # Status code mismatch will be used to trigger an error, but assertions will target
        # the sensitive fields redaction in the error message.
        start_config, next_config, end_config = self._create_custom_endpoint_configs(
            start_status_code=2000
        )

        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": CustomEndpointTarget(start_config, next_config, end_config),
            "stop_check": self._create_basic_stop_check(),
        }

        model_name = "Sensitive Fields Test " + rnd
        redacted_str = re.escape("*" * 16)
        exception_str_template = (
            "Response status code {status_code} did not match expected status code 2000. Response: .*. Request: [A-Z]+ https?://.*?, Headers: {'api-key': '"
            + redacted_str
            + "', 'Content-Type': 'application/json'}, Body: {}."
        )
        exception_str_match = exception_str_template.replace("{status_code}", "201")

        # Register the model with a custom endpoint that will fail.
        # Anticipate a redacted api-key
        self._register_and_expect_error(
            okareo,
            model_name,
            model_config,
            exception_str_match,
            run_test=True,
            basic_scenario=basic_scenario,
            api_keys={"openai": OPENAI_API_KEY},
            sensitive_fields=["api-key"],
        )

        # Register the model using redacted api-key field with same sensitive fields.
        # Anticipate same failure as above
        start_config, next_config, end_config = self._create_custom_endpoint_configs(
            start_status_code=2000, api_key="****************"
        )

        model_config = {
            "max_turns": 2,
            "repeats": 1,
            "target": CustomEndpointTarget(start_config, next_config, end_config),
            "stop_check": self._create_basic_stop_check(),
        }

        # Register the model with a custom endpoint that will fail.
        # Anticipate a redacted api-key
        self._register_and_expect_error(
            okareo,
            model_name,
            model_config,
            exception_str_match,
            run_test=True,
            basic_scenario=basic_scenario,
            api_keys={"openai": OPENAI_API_KEY},
            sensitive_fields=["api-key"],
        )

        # Register the model using redacted api-key field with api-key rm from sensitive fields.
        # Anticipate an auth error failure due to bad api-key
        exception_str_match = exception_str_template.replace("{status_code}", "401")
        self._register_and_expect_error(
            okareo,
            model_name,
            model_config,
            exception_str_match,
            run_test=True,
            basic_scenario=basic_scenario,
            api_keys={"openai": OPENAI_API_KEY},
            sensitive_fields=[],  # api-key should be overwritten with redacted string, causing 401
        )

    def test_nl_generation_missing_api_key(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when API keys are missing for NL_GENERATION test run"""
        target = self._create_basic_openai_target()
        model_name = f"Missing API Key NL Gen {rnd}"

        # Register the model
        mut = okareo.register_model(name=model_name, model=target, update=True)

        # Run test without providing api_keys
        with pytest.raises(Exception, match="Missing API"):
            mut.run_test(
                scenario=basic_scenario,
                name="Test Run",
                test_run_type=TestRunType.NL_GENERATION,
                checks=["coherence_summary"],
            )

    def test_nl_generation_no_checks(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when no checks are provided for NL_GENERATION test run"""
        target = self._create_basic_openai_target()
        model_name = f"No Checks NL Gen {rnd}"

        # Register the model
        mut = okareo.register_model(name=model_name, model=target, update=True)

        # Run test with empty checks list
        with pytest.raises(Exception, match="No checks were provided"):
            mut.run_test(
                scenario=basic_scenario,
                name="Test Run",
                test_run_type=TestRunType.NL_GENERATION,
                api_keys={"openai": OPENAI_API_KEY},
                checks=[],
            )

    def test_nl_generation_invalid_checks(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when no checks are provided for NL_GENERATION test run"""
        target = self._create_basic_openai_target()
        model_name = f"No Checks NL Gen {rnd}"

        # Register the model
        mut = okareo.register_model(name=model_name, model=target, update=True)

        # Run test with empty checks list
        with pytest.raises(
            Exception, match="One of more of the checks entered was invalid"
        ):
            mut.run_test(
                scenario=basic_scenario,
                name="Test Run",
                test_run_type=TestRunType.NL_GENERATION,
                api_keys={"openai": OPENAI_API_KEY},
                checks=["asdf"],
            )

    def test_custom_model_raises_value_error(self, rnd: str, okareo: Okareo) -> None:
        """Test failure when custom model raises a ValueError during invocation."""

        # Define a custom model that raises a ValueError
        class ErrorRaisingModel(CustomMultiturnTarget):
            def invoke(self, messages: list[dict[str, str]]) -> ModelInvocation:  # type: ignore
                # Always raise a ValueError with a specific message
                raise ValueError("This is a deliberate error from the custom model")

        # Create the model instance
        custom_model = ErrorRaisingModel(name=f"error_model_{rnd}")

        # Register the model with MultiTurnDriver
        model_under_test = okareo.register_model(
            name=f"Error Model Test {rnd}",
            model=MultiTurnDriver(
                driver_temperature=0.5,
                max_turns=3,
                repeats=1,
                target=custom_model,
                stop_check={"check_name": "model_refusal", "stop_on": False},
            ),
            update=True,
        )

        # Create a basic scenario
        seeds = [
            SeedData(
                input_="Hello, how are you today?",
                result="appropriate greeting response",
            ),
        ]

        scenario_set_create = ScenarioSetCreate(
            name=f"Error Model Test Scenario - {rnd}", seed_data=seeds
        )
        scenario = okareo.create_scenario_set(scenario_set_create)

        # Run the test and expect an error
        with pytest.raises(
            Exception, match="This is a deliberate error from the custom model"
        ):
            model_under_test.run_test(
                name=f"Error Model Test - {rnd}",
                api_key=OPENAI_API_KEY,
                scenario=scenario,
                test_run_type=TestRunType.MULTI_TURN,
                checks=["task_completed"],
            )

    def test_custom_model_parsing_error_from_return(
        self, rnd: str, okareo: Okareo
    ) -> None:
        """Test failure when custom model raises a ValueError during invocation."""

        # Define a custom model that raises a ValueError
        class ErrorRaisingModel(CustomMultiturnTarget):
            def invoke(self, messages: list[dict[str, str]]) -> Any:  # type: ignore
                # Always raise a ValueError with a specific message
                return "Not a modelinvocation object"

        # Create the model instance
        custom_model = ErrorRaisingModel(name=f"error_model_{rnd}")

        # Register the model with MultiTurnDriver
        model_under_test = okareo.register_model(
            name=f"Error Model Test {rnd}",
            model=MultiTurnDriver(
                driver_temperature=0.5,
                max_turns=3,
                repeats=1,
                target=custom_model,
                stop_check={"check_name": "model_refusal", "stop_on": False},
            ),
            update=True,
        )

        # Create a basic scenario
        seeds = [
            SeedData(
                input_="Hello, how are you today?",
                result="appropriate greeting response",
            ),
        ]

        scenario_set_create = ScenarioSetCreate(
            name=f"Error Model Test Scenario - {rnd}", seed_data=seeds
        )
        scenario = okareo.create_scenario_set(scenario_set_create)

        # Run the test and expect an error
        with pytest.raises(
            Exception,
            match="CustomModel did not return a response, or we were unable to parse it.",
        ):
            model_under_test.run_test(
                name=f"Error Model Test - {rnd}",
                api_key=OPENAI_API_KEY,
                scenario=scenario,
                test_run_type=TestRunType.MULTI_TURN,
                checks=[
                    "contains_all_imports",
                    "compression_ratio",
                    "are_all_params_expected",
                    "reverse_qa_quality",
                ],
            )

    def test_multiturn_error_checks(self, rnd: str, okareo: Okareo) -> None:
        """Test failure when invalid checks are provided for MULTI_TURN test run"""
        # Create a custom check that raises an error
        okareo.create_or_update_check(
            name="error_check",
            description="error_check",
            check=Check(),
        )

        # Create a basic multiturn model
        model_under_test = okareo.register_model(
            name=f"Error Check MT {rnd}",
            model=MultiTurnDriver(
                driver_temperature=0.5,
                max_turns=3,
                repeats=1,
                target=OpenAIModel(
                    model_id="gpt-4o-mini",
                    temperature=0,
                    system_prompt_template="test prompt",
                ),
                stop_check={"check_name": "model_refusal", "stop_on": False},
            ),
            update=True,
        )

        # Create a basic scenario
        seeds = [
            SeedData(
                input_="Test input",
                result="Test result",
            ),
        ]

        scenario_set_create = ScenarioSetCreate(
            name=f"Error Check Test Scenario - {rnd}", seed_data=seeds
        )
        scenario = okareo.create_scenario_set(scenario_set_create)

        # Run test with invalid check and expect error
        with pytest.raises(Exception, match="An error occurred while running checks"):
            model_under_test.run_test(
                scenario=scenario,
                name="Test Run",
                test_run_type=TestRunType.MULTI_TURN,
                api_keys={"openai": OPENAI_API_KEY},
                checks=["error_check"],
            )

    def test_openai_authentication_error(
        self, rnd: str, basic_scenario: Any, okareo: Any
    ) -> None:
        """Test failure when an invalid OpenAI API key is provided"""
        target = self._create_basic_openai_target()
        model_name = f"Invalid API Key Test {rnd}"

        # Register the model
        mut = okareo.register_model(name=model_name, model=target, update=True)

        # Run test with invalid API key and expect authentication error
        with pytest.raises(Exception, match="authentication"):
            mut.run_test(
                scenario=basic_scenario,
                name="Test Run",
                test_run_type=TestRunType.NL_GENERATION,
                api_keys={"openai": "test"},  # Invalid API key
                checks=["coherence_summary"],
            )



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.test_run_item_model_metrics import TestRunItemModelMetrics


T = TypeVar("T", bound="TestRunItem")


@_attrs_define
class TestRunItem:
    """
    Attributes:
        id (str):
        project_id (str):
        mut_id (Union[Unset, str]):
        scenario_set_id (Union[Unset, str]):
        filter_group_id (Union[Unset, str]):
        name (Union[Unset, str]):
        tags (Union[Unset, List[str]]):
        type (Union[Unset, str]):
        start_time (Union[Unset, datetime.datetime]):
        end_time (Union[Unset, None, datetime.datetime]):
        test_data_point_count (Union[Unset, int]):
        model_metrics (Union[Unset, TestRunItemModelMetrics]):
        error_matrix (Union[Unset, List[Any]]):
        status (Union[Unset, str]):
        failure_message (Union[Unset, str]):
        progress (Union[Unset, int]): Number in percent of progress of test run
        app_link (Union[Unset, str]): This URL links to the Okareo webpage for this test run Default: ''.
    """

    id: str
    project_id: str
    mut_id: Union[Unset, str] = UNSET
    scenario_set_id: Union[Unset, str] = UNSET
    filter_group_id: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    type: Union[Unset, str] = UNSET
    start_time: Union[Unset, datetime.datetime] = UNSET
    end_time: Union[Unset, None, datetime.datetime] = UNSET
    test_data_point_count: Union[Unset, int] = UNSET
    model_metrics: Union[Unset, "TestRunItemModelMetrics"] = UNSET
    error_matrix: Union[Unset, List[Any]] = UNSET
    status: Union[Unset, str] = UNSET
    failure_message: Union[Unset, str] = UNSET
    progress: Union[Unset, int] = 0
    app_link: Union[Unset, str] = ""
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        project_id = self.project_id
        mut_id = self.mut_id
        scenario_set_id = self.scenario_set_id
        filter_group_id = self.filter_group_id
        name = self.name
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        type = self.type
        start_time: Union[Unset, str] = UNSET
        if not isinstance(self.start_time, Unset):
            start_time = self.start_time.isoformat()

        end_time: Union[Unset, None, str] = UNSET
        if not isinstance(self.end_time, Unset):
            end_time = self.end_time.isoformat() if self.end_time else None

        test_data_point_count = self.test_data_point_count
        model_metrics: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.model_metrics, Unset):
            model_metrics = self.model_metrics.to_dict()

        error_matrix: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.error_matrix, Unset):
            error_matrix = self.error_matrix

        status = self.status
        failure_message = self.failure_message
        progress = self.progress
        app_link = self.app_link

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "project_id": project_id,
            }
        )
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if scenario_set_id is not UNSET:
            field_dict["scenario_set_id"] = scenario_set_id
        if filter_group_id is not UNSET:
            field_dict["filter_group_id"] = filter_group_id
        if name is not UNSET:
            field_dict["name"] = name
        if tags is not UNSET:
            field_dict["tags"] = tags
        if type is not UNSET:
            field_dict["type"] = type
        if start_time is not UNSET:
            field_dict["start_time"] = start_time
        if end_time is not UNSET:
            field_dict["end_time"] = end_time
        if test_data_point_count is not UNSET:
            field_dict["test_data_point_count"] = test_data_point_count
        if model_metrics is not UNSET:
            field_dict["model_metrics"] = model_metrics
        if error_matrix is not UNSET:
            field_dict["error_matrix"] = error_matrix
        if status is not UNSET:
            field_dict["status"] = status
        if failure_message is not UNSET:
            field_dict["failure_message"] = failure_message
        if progress is not UNSET:
            field_dict["progress"] = progress
        if app_link is not UNSET:
            field_dict["app_link"] = app_link

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.test_run_item_model_metrics import TestRunItemModelMetrics

        d = src_dict.copy()
        id = d.pop("id")

        project_id = d.pop("project_id")

        mut_id = d.pop("mut_id", UNSET)

        scenario_set_id = d.pop("scenario_set_id", UNSET)

        filter_group_id = d.pop("filter_group_id", UNSET)

        name = d.pop("name", UNSET)

        tags = cast(List[str], d.pop("tags", UNSET))

        type = d.pop("type", UNSET)

        _start_time = d.pop("start_time", UNSET)
        start_time: Union[Unset, datetime.datetime]
        if isinstance(_start_time, Unset):
            start_time = UNSET
        else:
            start_time = isoparse(_start_time)

        _end_time = d.pop("end_time", UNSET)
        end_time: Union[Unset, None, datetime.datetime]
        if _end_time is None:
            end_time = None
        elif isinstance(_end_time, Unset):
            end_time = UNSET
        else:
            end_time = isoparse(_end_time)

        test_data_point_count = d.pop("test_data_point_count", UNSET)

        _model_metrics = d.pop("model_metrics", UNSET)
        model_metrics: Union[Unset, TestRunItemModelMetrics]
        if isinstance(_model_metrics, Unset):
            model_metrics = UNSET
        else:
            model_metrics = TestRunItemModelMetrics.from_dict(_model_metrics)

        error_matrix = cast(List[Any], d.pop("error_matrix", UNSET))

        status = d.pop("status", UNSET)

        failure_message = d.pop("failure_message", UNSET)

        progress = d.pop("progress", UNSET)

        app_link = d.pop("app_link", UNSET)

        test_run_item = cls(
            id=id,
            project_id=project_id,
            mut_id=mut_id,
            scenario_set_id=scenario_set_id,
            filter_group_id=filter_group_id,
            name=name,
            tags=tags,
            type=type,
            start_time=start_time,
            end_time=end_time,
            test_data_point_count=test_data_point_count,
            model_metrics=model_metrics,
            error_matrix=error_matrix,
            status=status,
            failure_message=failure_message,
            progress=progress,
            app_link=app_link,
        )

        test_run_item.additional_properties = d
        return test_run_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from okareo.checks import CodeBasedCheck


class Check(CodeBasedCheck):
    @staticmethod
    def evaluate(model_output: str, scenario_input: str, scenario_result: str, metadata: dict, model_input: str) -> int:  # fmt: skip
        model_input_len = sum([len(str(mi)) for mi in model_input])
        return model_input_len > len(scenario_input)



from okareo.checks import CodeBasedCheck


class Check(CodeBasedCheck):
    @staticmethod
    def evaluate(model_output: str, scenario_input: str, scenario_result: str, metadata: dict, model_input: str) -> bool:  # fmt: skip
        # Check if metadata is empty
        if metadata.get("tool_calls"):
            return True if len(metadata["tool_calls"]) > 0 else False
        return False
